# sort the dependencies

import copy
from typing import List


    
def adjust_resolved_dependencies(avro_schema: List[dict] | dict):
    """
    After resolving dependencies, it may still be necessary to adjust them. The
    first pass of the algorithms below does inline all dependent types, but 
    the resulting document may still have fields defined before the types they
    depend on because of the order in which the resolution happened, which necessarily 
    re-sorts the graph. This function will recursively adjust the resolved
    dependencies until all record types have their dependency types defined before them.
    """

    class TreeWalker:

        def __init__(self):
            self.found_something = True

        def swap_record_dependencies_above(self, current_node, record, avro_schema) -> str | None:
            """ swap the first reference to of the record type above the record in avro_schema """
            if isinstance(current_node, dict):
                if 'name' in current_node and 'namespace' in current_node and 'type' in current_node and \
                    current_node['name'] == record['name'] and current_node.get('namespace','') == record.get('namespace','') and current_node['type'] == record['type']:
                    # we reached the record again. we stop here.
                    return None
                for k, v in current_node.items():
                    if k in ['dependencies', 'unmerged_types']:
                        continue
                    if isinstance(v, (dict,list)):
                        return self.swap_record_dependencies_above(v, record, avro_schema)
                    elif isinstance(v, str):
                        if k not in ['type', 'values', 'items']:
                            continue
                        qname = record.get('namespace','')+'.'+record['name']
                        if v == qname:
                            self.found_something = True
                            current_node[k] = copy.deepcopy(record)
                            return qname
            elif isinstance(current_node, list):
                for item in current_node:
                    if isinstance(item, (dict,list)):
                        return self.swap_record_dependencies_above(item, record, avro_schema)
                    elif isinstance(item, str):
                        qname = record.get('namespace','')+'.'+record['name']
                        if item == qname:
                            self.found_something = True
                            idx = current_node.index(item)
                            current_node.remove(item)
                            current_node.insert(idx, copy.deepcopy(record))
                            return qname
            return None
        
        def walk_schema(self, current_node, avro_schema, record_list) -> str | None:
            found_record = None
            if isinstance(current_node, dict):
                if 'type' in current_node and (current_node['type'] == 'record' or current_node['type'] == 'enum'):
                    current_qname = current_node.get('namespace','')+'.'+current_node.get('name','')
                    if current_qname in record_list:
                        self.found_something = True
                        return current_qname
                    record_list.append(current_qname)
                    found_record = self.swap_record_dependencies_above(avro_schema, current_node, avro_schema)
                for k, v in current_node.items():
                    if isinstance(v, (dict,list)):
                        qname = self.walk_schema(v, avro_schema, record_list)
                        if qname:
                            self.found_something = True
                            current_node[k] = qname
            elif isinstance(current_node, list):
                for item in current_node:
                    qname = self.walk_schema(item, avro_schema, record_list)
                    if qname:
                        self.found_something = True
                        idx = current_node.index(item)
                        current_node.remove(item)
                        current_node.insert(idx, qname)
                # dedupe the list
                new_list = []
                for item in current_node:
                    if not item in new_list:
                        new_list.append(item)
                current_node.clear()
                current_node.extend(new_list)
            return found_record

    # while we've got work to do
    tree_walker = TreeWalker()
    while True:
        tree_walker.found_something = False
        tree_walker.walk_schema(avro_schema, avro_schema, [])
        if not tree_walker.found_something:
            break



def inline_dependencies_of(avro_schema, record):
    """ to break circular dependencies, we will inline all dependent record """
    for dependency in copy.deepcopy(record.get('dependencies', [])):
        dependency_type = next((x for x in avro_schema if x['name'] == dependency or x.get('namespace','')+'.'+x['name'] == dependency), None)
        if not dependency_type:
            continue
        deps = record.get('dependencies', [])
        for field in record['fields']:                        
            swap_dependency_type(avro_schema, field, dependency, dependency_type, deps, [record['namespace']+'.'+record['name']])
    if 'dependencies' in record:
        del record['dependencies']

    adjust_resolved_dependencies(record)

    

def sort_messages_by_dependencies(avro_schema):
    """ 
        Sort the messages in avro_schema by their dependencies. Avro Schema requires
        that type definitions must be defined before they are used. This method
        ensures this. Types that have dependencies will be moved at the end of the list.
        If necessary, it will also resolve circular dependencies by inlining the 
        dependent record.

        The method expects all types with dependencies to have a 'dependencies' key in their
        dict that contains a list of types that they depend on.

        Args:
            avro_schema: List of Avro schema records.
    """

    # if all are just strings, then it is already sorted
    if all(isinstance(record, str) for record in avro_schema):
        return avro_schema

    sorted_messages = []
    record_stack = []
    while avro_schema:
        found = False
        for record in avro_schema:
            if not isinstance(record, dict):
                sorted_messages.append(record)
                avro_schema.remove(record)
                continue
            
            # if this record is not a dependency of any other record, it can be safely emitted now
            #if not any(record.get('namespace','')+'.'+record.get('name') in other_record.get('dependencies', []) for other_record in [x for x in avro_schema if isinstance(x, dict) and 'name' in x]):
            remaining_deps = [dep for dep in record['dependencies'] if not dep in [x.get('namespace','')+'.'+x.get('name','') for x in sorted_messages]] if 'dependencies' in record else []
            if len(remaining_deps) == 0:
                if 'dependencies' in record:
                    del record['dependencies']
                sorted_messages.append(record)
                avro_schema.remove(record)
                found = True
                
        # If there are no records without dependencies, we will grab the first 
        # record with dependencies and start resolving circular dependencies
        if len(avro_schema) > 0 and not found:
            found = False
            for record in avro_schema:
                if isinstance(record, dict) and 'dependencies' in record:
                    remaining_deps = [dep for dep in record['dependencies'] if not dep in [x.get('namespace','')+'.'+x.get('name','') for x in sorted_messages]]
                    if len(remaining_deps) > 0:
                        swap_record_dependencies(avro_schema, record, [record.get('namespace','')+'.'+record['name']], 0)
                        if 'dependencies' in record and len(record['dependencies']) == 0:
                            del record['dependencies']
                        if isinstance(record, dict) and not 'dependencies' in record:
                            found = True
                            sorted_messages.append(record)
                            if record in avro_schema:
                                avro_schema.remove(record)
                            break
                        else:
                            remaining_remaining_deps = [dep for dep in record['dependencies'] if not dep in [x.get('namespace')+'.'+x.get('name') for x in sorted_messages]]
                            found = len(remaining_deps) != len(remaining_remaining_deps)
                            if found:
                                break
                            
            if not found:
                found = False
                for record in avro_schema:
                    if isinstance(record, dict) and 'dependencies' in record:            
                        found = True
                        record_deps = copy.deepcopy(record.get('dependencies', []))
                        inline_dependencies_of(avro_schema, record)
                        # fix the dependencies of all records that have this record as a dependency
                        for remaining_schema in avro_schema:
                            if isinstance(remaining_schema, dict) and 'dependencies' in remaining_schema and any(dep in record_deps for dep in remaining_schema['dependencies']):
                                remaining_schema['dependencies'] = [dep for dep in remaining_schema['dependencies'] if not dep in record_deps]
                                qname = record['namespace']+'.'+record['name']
                                if not qname in remaining_schema['dependencies']:
                                    remaining_schema['dependencies'].append(qname)
                        break

                if not found:
                    print('WARNING: There are circular dependencies in the schema, unable to resolve them: {}'.format([x['name'] for x in avro_schema if isinstance(x, dict) and 'dependencies' in x]))
    
    adjust_resolved_dependencies(sorted_messages)
    return sorted_messages

def swap_record_dependencies(avro_schema, record, record_stack: List[str], recursion_depth: int = 0):
    record_stack.append(record.get('namespace', '')+'.'+record['name'])
    if 'dependencies' in record:
        prior_dependencies = copy.deepcopy(record['dependencies'])
        while 'dependencies' in record and len(record['dependencies']) > 0:
            if 'fields' in record:
                for field in record['fields']:
                    if isinstance(field['type'], list):
                        for item in field['type'].copy():
                            sub_field = {
                                'type': item,
                                'name': field['name']   
                            }
                            resolve_field_dependencies(avro_schema, record, sub_field, record_stack, recursion_depth + 1)
                            if sub_field['type'] != item:
                                idx = field['type'].index(item)
                                field['type'].remove(item)
                                field['type'].insert(idx, sub_field['type'])
                    else:
                        resolve_field_dependencies(avro_schema, record, field, record_stack, recursion_depth + 1)
            if 'dependencies' in record and len(record['dependencies']) > 0:
                # compare the prior dependencies to the current dependencies one-by-one. If they are the same,
                # then we have a circular dependency.
                if prior_dependencies == record['dependencies']:
                    print('WARNING: Unable to resolve circular dependency in {}::{} with dependencies: {}'.format(record.get('namespace',''), record['name'], record['dependencies']))
                    break
                prior_dependencies = record['dependencies']
        if 'dependencies' in record:
            del record['dependencies']
    record_stack.pop()

def resolve_field_dependencies(avro_schema, record, field, record_stack, recursion_depth: int = 0):
    for dependency in record.get('dependencies', []):
        dependency_type = next((x for x in avro_schema if x['name'] == dependency or x.get('namespace','')+'.'+x['name'] == dependency), None)
        if not dependency_type and dependency in record['dependencies']:
            record['dependencies'].remove(dependency)
            continue
        deps = record.get('dependencies', [])
        if dependency_type:
            if record['name'] != dependency and (record.get('namespace','')+'.'+record['name']) != dependency:
                swap_dependency_type(avro_schema, field, dependency, dependency_type, deps, record_stack, recursion_depth + 1)
        record['dependencies'] = [dep for dep in deps if dep != record['name'] and record.get('namespace','')+'.'+record['name'] != dep]
        if len(record['dependencies']) == 0:
            del record['dependencies']


def swap_dependency_type(avro_schema, field, dependency, dependency_type, dependencies, record_stack: List[str], recursion_depth: int = 0):
    """ to break circular dependencies, we will inline the dependent record and remove the dependency """
    if not dependency in dependencies:
        return
    if not dependency_type in avro_schema:
        return
    if record_stack and dependency in record_stack:
        dependencies.remove(dependency)
        return
    
    # Replace the dependency type with the dependency_type in avro_schema.
    if isinstance(field['type'],str) and field['type'] == dependency:
        if dependency_type in avro_schema:
            field['type'] = dependency_type
            avro_schema.remove(dependency_type)
        dependencies.remove(dependency)
        dependencies.extend(dependency_type.get('dependencies', []))
        if 'dependencies' in dependency_type:
            swap_record_dependencies(avro_schema, dependency_type, record_stack, recursion_depth + 1)
           
    # type is a Union?
    elif isinstance(field['type'], list):
        for field_type in field['type']:
            if field_type == dependency:
                if dependency_type in avro_schema:
                    index = field['type'].index(field_type)
                    field['type'].remove(field_type)
                    field['type'].insert(index, dependency_type)
                    avro_schema.remove(dependency_type)
                if dependency in dependencies:
                    dependencies.remove(dependency)
                dependencies.extend(dependency_type.get('dependencies', []))
                if 'dependencies' in dependency_type:
                    swap_record_dependencies(avro_schema, dependency_type, record_stack, recursion_depth + 1)
        for field_type in field['type']:
            if isinstance(field_type, dict):
                swap_dependency_type(avro_schema, field_type, dependency, dependency_type, dependencies, record_stack, recursion_depth + 1)
    elif isinstance(field['type'], dict) and 'type' in field['type']:
        swap_dependency_type(avro_schema, field['type'], dependency, dependency_type, dependencies, record_stack, recursion_depth + 1)
    elif field['type'] == 'array':
        if not 'items' in field:
            return
        if isinstance(field['items'], list):
            for item in field['items']:
                if item == dependency:
                    if dependency_type in avro_schema:
                        index = field['items'].index(item)
                        field['items'].remove(item)
                        field['items'].insert(index, dependency_type)
                        avro_schema.remove(dependency_type)
                    if dependency in dependencies:
                        dependencies.remove(dependency)
                    dependencies.extend(dependency_type.get('dependencies', []))
                    if 'dependencies' in dependency_type:
                        swap_record_dependencies(avro_schema, dependency_type, record_stack)
            for item in field['items']:
                if isinstance(item, dict):
                    swap_dependency_type(avro_schema, item, dependency, dependency_type, dependencies, record_stack, recursion_depth + 1)               
        elif field['items'] == dependency:
            if dependency_type in avro_schema:
                field['items'] = dependency_type
                avro_schema.remove(dependency_type)
            if dependency in dependencies:
                dependencies.remove(dependency)
            dependencies.extend(dependency_type.get('dependencies', []))
            if 'dependencies' in dependency_type:
                swap_record_dependencies(avro_schema, dependency_type, record_stack)
        elif isinstance(field['items'], dict) and 'type' in field['items']:
            swap_dependency_type(avro_schema, field['items'], dependency, dependency_type, dependencies, record_stack, recursion_depth + 1)
    elif field['type'] == 'map':
        if isinstance(field['values'], list):
            for item in field['values']:
                if item == dependency:
                    if dependency_type in avro_schema:
                        index = field['values'].index(item)
                        field['values'].remove(item)
                        field['values'].insert(index, dependency_type)
                        avro_schema.remove(dependency_type)
                    if dependency in dependencies:
                        dependencies.remove(dependency)
                    dependencies.extend(dependency_type.get('dependencies', []))
                    if 'dependencies' in dependency_type:
                        swap_record_dependencies(avro_schema, dependency_type, record_stack)
            for item in field['values']:
                if isinstance(item, dict):
                    swap_dependency_type(avro_schema, item, dependency, dependency_type, dependencies, record_stack, recursion_depth + 1)
        if field['values'] == dependency:
            if dependency_type in avro_schema:
                field['values'] = dependency_type
                avro_schema.remove(dependency_type) 
            if dependency in dependencies:                   
                dependencies.remove(dependency)
            dependencies.extend(dependency_type.get('dependencies', []))
            if 'dependencies' in dependency_type:
                swap_record_dependencies(avro_schema, dependency_type, record_stack)
        elif 'type' in field['values']:
            swap_dependency_type(avro_schema, field['values'], dependency, dependency_type, dependencies, record_stack, recursion_depth + 1)
    elif field['type'] == 'record':
        record_stack.append(field.get('namespace', '')+'.'+field['name'])
        for dep_field in field['fields']:
            if isinstance(dep_field, dict):
                swap_dependency_type(avro_schema, dep_field, dependency, dependency_type, dependencies, record_stack, recursion_depth + 1)  
        record_stack.pop()



# coding: utf-8
"""
Module to convert JSON Structure schema to Markdown documentation.
"""

import json
import os
from typing import Any, Dict, List, Union

from avrotize.common import render_template

class StructureToMarkdownConverter:
    """
    Class to convert JSON Structure schema to Markdown documentation.
    """

    def __init__(self, structure_schema_path: str, markdown_path: str):
        """
        Initialize the converter with file paths.

        :param structure_schema_path: Path to the JSON Structure schema file.
        :param markdown_path: Path to save the Markdown file.
        """
        self.structure_schema_path = structure_schema_path
        self.markdown_path = markdown_path
        self.objects = {}
        self.choices = {}
        self.enums = {}
        self.definitions = {}

    def convert(self):
        """
        Convert JSON Structure schema to Markdown and save to file.
        """
        with open(self.structure_schema_path, 'r', encoding='utf-8') as file:
            structure_schema = json.load(file)

        schema_name = os.path.splitext(os.path.basename(self.structure_schema_path))[0]
        
        # Process the schema
        self.process_schema(structure_schema)
        
        # Generate markdown
        self.generate_markdown(schema_name, structure_schema)

    def process_schema(self, schema: Dict[str, Any], parent_namespace: str = ''):
        """
        Process the schema to extract all named types.
        
        :param schema: The schema or sub-schema to process
        :param parent_namespace: The parent namespace
        """
        # Store definitions separately
        if 'definitions' in schema:
            self.definitions = schema['definitions']
        
        # Process root schema (but not definitions)
        if 'type' in schema:
            name = schema.get('name', 'Root')
            self.extract_named_types(schema, parent_namespace, name, skip_definitions=True)

    def extract_named_types(self, schema: Union[Dict, List, str], parent_namespace: str = '', type_name: str = '', skip_definitions: bool = False):
        """
        Extract all named types (objects, choices, enums) from the schema.
        
        :param schema: Schema element to process
        :param parent_namespace: Parent namespace
        :param type_name: Name of this type
        :param skip_definitions: Skip processing definitions (they're handled separately)
        """
        if isinstance(schema, dict):
            # Skip definitions section when processing root schema
            if skip_definitions and 'definitions' in schema:
                # Create a copy without definitions for processing
                schema = {k: v for k, v in schema.items() if k != 'definitions'}
            
            schema_type = schema.get('type')
            ns = schema.get('namespace', parent_namespace)
            name = schema.get('name', type_name)
            
            # Only add named objects and choices to the respective lists
            # Inline/anonymous objects should not be listed separately
            if schema_type == 'object' and 'name' in schema:
                self.objects.setdefault(ns, []).append(schema)
            elif schema_type == 'choice' and 'name' in schema:
                self.choices.setdefault(ns, []).append(schema)
            elif 'enum' in schema:
                # Handle enum constraint
                self.enums.setdefault(ns, []).append({
                    'name': name,
                    'values': schema['enum'],
                    'doc': schema.get('description', '')
                })
            
            # Recursively process properties
            if 'properties' in schema:
                for prop_name, prop_schema in schema['properties'].items():
                    self.extract_named_types(prop_schema, ns, prop_name)
            
            # Process items in arrays/sets
            if 'items' in schema:
                self.extract_named_types(schema['items'], ns)
            
            # Process values in maps
            if 'values' in schema:
                self.extract_named_types(schema['values'], ns)
            
            # Process choice alternatives
            if 'choices' in schema:
                choices = schema['choices']
                if isinstance(choices, dict):
                    for choice_name, choice_schema in choices.items():
                        self.extract_named_types(choice_schema, ns, choice_name)
                elif isinstance(choices, list):
                    for choice_schema in choices:
                        self.extract_named_types(choice_schema, ns)
                        
        elif isinstance(schema, list):
            # Union type
            for sub_schema in schema:
                self.extract_named_types(sub_schema, parent_namespace)

    def generate_markdown(self, schema_name: str, root_schema: Dict[str, Any]):
        """
        Generate markdown content from the extracted types using Jinja2 template.

        :param schema_name: The name of the schema file.
        :param root_schema: The root schema object
        """
        render_template("structuretomd/README.md.jinja", self.markdown_path,
            schema_name=schema_name,
            schema_description=root_schema.get('description', ''),
            schema_id=root_schema.get('$id', ''),
            schema_uses=root_schema.get('$uses', []),
            schema_offers=root_schema.get('$offers', {}),
            objects=self.objects,
            choices=self.choices,
            enums=self.enums,
            definitions=self.definitions,
            generate_property_markdown=self.generate_property_markdown,
            get_type_string=self.get_type_string
        )

    def generate_property_markdown(self, prop_name: str, prop_schema: Union[Dict, List, str], required: bool = False) -> str:
        """
        Generate markdown content for a single property.

        :param prop_name: Property name
        :param prop_schema: Property schema
        :param required: Whether the property is required
        :return: Markdown content as a string
        """
        lines = []
        type_str = self.get_type_string(prop_schema)
        required_str = " (required)" if required else ""
        
        lines.append(f"- **{prop_name}**{required_str}: {type_str}")
        
        # Add description if present
        if isinstance(prop_schema, dict):
            if 'description' in prop_schema:
                lines.append(f"  - Description: {prop_schema['description']}")
            
            # Add extension properties (from various JSON Structure extension specs)
            extensions = []
            
            # Alternate names (JSONStructureAlternateNames extension)
            if 'altnames' in prop_schema:
                altnames = prop_schema['altnames']
                if isinstance(altnames, dict):
                    altnames_list = [f"{k}: {v}" for k, v in altnames.items()]
                    extensions.append(f"altnames: {{{', '.join(altnames_list)}}}")
            
            # Units (JSONStructureUnits extension)
            if 'unit' in prop_schema:
                extensions.append(f"unit: {prop_schema['unit']}")
            
            # Currency (for decimal types)
            if 'currency' in prop_schema:
                extensions.append(f"currency: {prop_schema['currency']}")
            
            # Symbol (from JSONStructureSymbol extension)
            if 'symbol' in prop_schema:
                extensions.append(f"symbol: {prop_schema['symbol']}")
            
            # Content encoding/media type (JSONStructureContent extension)
            if 'contentEncoding' in prop_schema:
                extensions.append(f"contentEncoding: {prop_schema['contentEncoding']}")
            if 'contentMediaType' in prop_schema:
                extensions.append(f"contentMediaType: {prop_schema['contentMediaType']}")
            
            # Format
            if 'format' in prop_schema:
                extensions.append(f"format: {prop_schema['format']}")
            
            # Examples
            if 'examples' in prop_schema:
                examples_str = ', '.join(str(ex) for ex in prop_schema['examples'])
                extensions.append(f"examples: [{examples_str}]")
            
            # Default value
            if 'default' in prop_schema:
                extensions.append(f"default: {prop_schema['default']}")
            
            if extensions:
                lines.append(f"  - Extensions: {', '.join(extensions)}")
            
            # Add constraints
            constraints = []
            if 'minLength' in prop_schema:
                constraints.append(f"minLength: {prop_schema['minLength']}")
            if 'maxLength' in prop_schema:
                constraints.append(f"maxLength: {prop_schema['maxLength']}")
            if 'minimum' in prop_schema:
                constraints.append(f"minimum: {prop_schema['minimum']}")
            if 'maximum' in prop_schema:
                constraints.append(f"maximum: {prop_schema['maximum']}")
            if 'exclusiveMinimum' in prop_schema:
                constraints.append(f"exclusiveMinimum: {prop_schema['exclusiveMinimum']}")
            if 'exclusiveMaximum' in prop_schema:
                constraints.append(f"exclusiveMaximum: {prop_schema['exclusiveMaximum']}")
            if 'multipleOf' in prop_schema:
                constraints.append(f"multipleOf: {prop_schema['multipleOf']}")
            if 'pattern' in prop_schema:
                constraints.append(f"pattern: `{prop_schema['pattern']}`")
            if 'minItems' in prop_schema:
                constraints.append(f"minItems: {prop_schema['minItems']}")
            if 'maxItems' in prop_schema:
                constraints.append(f"maxItems: {prop_schema['maxItems']}")
            if 'precision' in prop_schema:
                constraints.append(f"precision: {prop_schema['precision']}")
            if 'scale' in prop_schema:
                constraints.append(f"scale: {prop_schema['scale']}")
            if 'enum' in prop_schema:
                constraints.append(f"enum: {', '.join(str(v) for v in prop_schema['enum'])}")
            if 'const' in prop_schema:
                constraints.append(f"const: {prop_schema['const']}")
            
            if constraints:
                lines.append(f"  - Constraints: {', '.join(constraints)}")
        
        return '\n'.join(lines)

    def get_type_string(self, schema: Union[Dict, List, str]) -> str:
        """
        Get a string representation of a type.

        :param schema: Type schema
        :return: Type string
        """
        if isinstance(schema, str):
            # Simple type reference or primitive
            if schema.startswith('#/'):
                # Reference
                ref_parts = schema.split('/')
                return f"[{ref_parts[-1]}](#{ref_parts[-1].lower()})"
            return f"`{schema}`"
        
        elif isinstance(schema, list):
            # Union type
            type_strs = [self.get_type_string(t) for t in schema]
            return ' | '.join(type_strs)
        
        elif isinstance(schema, dict):
            schema_type = schema.get('type')
            
            if '$ref' in schema:
                # Reference
                ref = schema['$ref']
                ref_parts = ref.split('/')
                ref_name = ref_parts[-1]
                return f"[{ref_name}](#{ref_name.lower()})"
            
            if schema_type == 'array':
                items_type = self.get_type_string(schema.get('items', 'any'))
                return f"array&lt;{items_type}&gt;"
            
            elif schema_type == 'set':
                items_type = self.get_type_string(schema.get('items', 'any'))
                return f"set&lt;{items_type}&gt;"
            
            elif schema_type == 'map':
                keys_type = self.get_type_string(schema.get('keys', 'string'))
                values_type = self.get_type_string(schema.get('values', 'any'))
                return f"map&lt;{keys_type}, {values_type}&gt;"
            
            elif schema_type == 'tuple':
                items = schema.get('items', [])
                if isinstance(items, list):
                    item_types = [self.get_type_string(item) for item in items]
                    return f"tuple&lt;{', '.join(item_types)}&gt;"
                return "tuple"
            
            elif schema_type == 'choice':
                name = schema.get('name', 'Choice')
                return f"[{name}](#{name.lower()})"
            
            elif schema_type == 'object':
                name = schema.get('name', 'object')
                if name != 'object':
                    return f"[{name}](#{name.lower()})"
                return "`object`"
            
            elif schema_type:
                return f"`{schema_type}`"
            
            # If no type specified but has properties, it's an inline object
            if 'properties' in schema:
                return "`object` (inline)"
        
        return "`any`"


def convert_structure_to_markdown(structure_schema_path: str, markdown_path: str):
    """
    Convert a JSON Structure schema file to a Markdown file.

    :param structure_schema_path: Path to the JSON Structure schema file.
    :param markdown_path: Path to save the Markdown file.
    """
    converter = StructureToMarkdownConverter(structure_schema_path, markdown_path)
    converter.convert()



import importlib

mod = "avrotize"
class LazyLoader:
    """    
    Lazy loader for the avrotize functions to speed up startup time.    
    """
    def __init__(self, mappings):
        self._modules = {}
        self._mappings = mappings

    def _load_module(self, module_name):
        if module_name not in self._modules:
            self._modules[module_name] = importlib.import_module(module_name)
        return self._modules[module_name]

    def __getattr__(self, item):
        if item in self._mappings:
            module_name, func_name = self._mappings[item]
            module = self._load_module(module_name)
            return getattr(module, func_name)
        else:
            return self._load_module(f"{mod}.{item}")

# Define the functions and their corresponding module paths
_mappings = {
    "convert_proto_to_avro": (f"{mod}.prototoavro", "convert_proto_to_avro"),
    "convert_jsons_to_avro": (f"{mod}.jsonstoavro", "convert_jsons_to_avro"),
    "convert_kafka_struct_to_avro_schema": (f"{mod}.kstructtoavro", "convert_kafka_struct_to_avro_schema"),
    "convert_kusto_to_avro": (f"{mod}.kustotoavro", "convert_kusto_to_avro"),
    "convert_asn1_to_avro": (f"{mod}.asn1toavro", "convert_asn1_to_avro"),
    "convert_xsd_to_avro": (f"{mod}.xsdtoavro", "convert_xsd_to_avro"),
    "convert_avro_to_json_schema": (f"{mod}.avrotojsons", "convert_avro_to_json_schema"),
    "convert_avro_to_kusto_file": (f"{mod}.avrotokusto", "convert_avro_to_kusto_file"),
    "convert_avro_to_kusto_db": (f"{mod}.avrotokusto", "convert_avro_to_kusto_db"),
    "convert_avro_to_parquet": (f"{mod}.avrotoparquet", "convert_avro_to_parquet"),
    "convert_avro_to_proto": (f"{mod}.avrotoproto", "convert_avro_to_proto"),
    "convert_avro_to_sql": (f"{mod}.avrotodb", "convert_avro_to_sql"),
    "convert_avro_to_xsd": (f"{mod}.avrotoxsd", "convert_avro_to_xsd"),
    "convert_avro_to_java": (f"{mod}.avrotojava", "convert_avro_to_java"),
    "convert_avro_schema_to_java": (f"{mod}.avrotojava", "convert_avro_schema_to_java"),
    "convert_avro_to_csharp": (f"{mod}.avrotocsharp", "convert_avro_to_csharp"),
    "convert_avro_schema_to_csharp": (f"{mod}.avrotocsharp", "convert_avro_schema_to_csharp"),
    "convert_avro_to_python": (f"{mod}.avrotopython", "convert_avro_to_python"),
    "convert_avro_schema_to_python": (f"{mod}.avrotopython", "convert_avro_schema_to_python"),
    "convert_avro_to_typescript": (f"{mod}.avrotots", "convert_avro_to_typescript"),
    "convert_avro_schema_to_typescript": (f"{mod}.avrotots", "convert_avro_schema_to_typescript"),
    "convert_avro_to_javascript": (f"{mod}.avrotojs", "convert_avro_to_javascript"),
    "convert_avro_schema_to_javascript": (f"{mod}.avrotojs", "convert_avro_schema_to_javascript"),
    "convert_avro_to_markdown": (f"{mod}.avrotomd", "convert_avro_to_markdown"),
    "convert_structure_to_markdown": (f"{mod}.structuretomd", "convert_structure_to_markdown"),
    "convert_avro_to_cpp": (f"{mod}.avrotocpp", "convert_avro_to_cpp"),
    "convert_avro_schema_to_cpp": (f"{mod}.avrotocpp", "convert_avro_schema_to_cpp"),
    "convert_avro_to_go": (f"{mod}.avrotogo", "convert_avro_to_go"),
    "convert_avro_schema_to_go": (f"{mod}.avrotogo", "convert_avro_schema_to_go"),
    "convert_avro_to_rust": (f"{mod}.avrotorust", "convert_avro_to_rust"),
    "convert_avro_schema_to_rust": (f"{mod}.avrotorust", "convert_avro_schema_to_rust"),
    "convert_avro_to_datapackage": (f"{mod}.avrotodatapackage", "convert_avro_to_datapackage"),
    "convert_structure_to_javascript": (f"{mod}.structuretojs", "convert_structure_to_javascript"),
    "convert_structure_schema_to_javascript": (f"{mod}.structuretojs", "convert_structure_schema_to_javascript"),
}

_lazy_loader = LazyLoader(_mappings)

def __getattr__(name):
    return getattr(_lazy_loader, name)



"""Converts Avro schema to Python data classes"""

# pylint: disable=line-too-long,too-many-instance-attributes

import copy
import json
import os
import re
import random
from typing import Dict, List, Set, Tuple, Union, Any
from avrotize.common import fullname, get_typing_args_from_string, is_generic_avro_type, pascal, process_template, build_flat_type_dict, inline_avro_references, is_type_with_alternate, strip_alternate_type

INDENT = '    '


def is_python_reserved_word(word: str) -> bool:
    """Checks if a word is a Python reserved word"""
    reserved_words = [
        'False', 'None', 'True', 'and', 'as', 'assert', 'async', 'await',
        'break', 'class', 'continue', 'def', 'del', 'elif', 'else', 'except',
        'finally', 'for', 'from', 'global', 'if', 'import', 'in', 'is',
        'lambda', 'nonlocal', 'not', 'or', 'pass', 'raise', 'return',
        'try', 'while', 'with', 'yield', 'record', 'self', 'cls'
    ]
    return word in reserved_words


class AvroToPython:
    """Converts Avro schema to Python data classes"""

    def __init__(self, base_package: str = '', dataclasses_json_annotation=False, avro_annotation=False) -> None:
        self.base_package = base_package
        self.dataclasses_json_annotation = dataclasses_json_annotation
        self.avro_annotation = avro_annotation
        self.output_dir = os.getcwd()
        self.main_schema = None
        self.type_dict = None
        self.generated_types: Dict[str, str] = {}

    def is_python_primitive(self, type_name: str) -> bool:
        """ Checks if a type is a Python primitive type """
        return type_name in ['None', 'bool', 'int', 'float', 'str', 'bytes']

    def is_python_typing_struct(self, type_name: str) -> bool:
        """ Checks if a type is a Python typing type """
        return type_name.startswith('typing.Dict[') or type_name.startswith('typing.List[') or type_name.startswith('typing.Optional[') or type_name.startswith('typing.Union[') or type_name == 'typing.Any'

    def safe_name(self, name: str) -> str:
        """Converts a name to a safe Python name"""
        if is_python_reserved_word(name):
            return name + "_"
        return name

    def pascal_type_name(self, ref: str) -> str:
        """Converts a reference to a type name"""
        return '_'.join([pascal(part) for part in ref.split('.')[-1].split('_')])

    def python_package_from_avro_type(self, namespace: str, type_name: str) -> str:
        """Gets the Python package from a type name"""
        type_name_package = '.'.join([part.lower() for part in type_name.split('.')]) if '.' in type_name else type_name.lower()
        if '.' in type_name:
            #  if the type name was already qualified, we don't need to add the namespace
            package = type_name_package
        else:
            namespace_package = '.'.join([part.lower() for part in namespace.split('.')]) if namespace else ''
            package = namespace_package + ('.' if namespace_package and type_name_package else '') + type_name_package
        if self.base_package:
            package = self.base_package + '.' + package
        return package

    def python_type_from_avro_type(self, type_name: str) -> str:
        """Gets the Python class from a type name"""
        return self.pascal_type_name(type_name)

    def python_fully_qualified_name_from_avro_type(self, namespace: str, type_name: str) -> str:
        """
        Gets the fully qualified Python class name from an Avro type.
        """
        package = self.python_package_from_avro_type(namespace, type_name)
        return package + ('.' if package else '') + self.python_type_from_avro_type(type_name)

    def strip_package_from_fully_qualified_name(self, fully_qualified_name: str) -> str:
        """Strips the package from a fully qualified name"""
        return fully_qualified_name.split('.')[-1]

    def map_plain_type_reference_to_python(self, parent_namespace: str, avro_type: str) -> Tuple[bool, str]:
        """
        Maps an Avro type to a Python type

        Args:
            avro_type (str): Avro type

        Returns:
            Tuple[bool, str]: A tuple containing a boolean indicating
                if the type is a primitive type and the Python type
        """
        mapping = {
            'null': 'None',
            'boolean': 'bool',
            'int': 'int',
            'long': 'int',
            'float': 'float',
            'double': 'float',
            'bytes': 'bytes',
            'string': 'str',
        }
        if is_generic_avro_type(avro_type):
            return True, 'typing.Any'
        mapped = mapping.get(avro_type, None)
        if mapped:
            return True, mapped
        return False, self.python_fully_qualified_name_from_avro_type(parent_namespace, avro_type)

    def convert_logical_type_to_python(self, avro_type: Dict, import_types: Set[str]) -> str:
        """Converts Avro logical type to Python type"""
        if avro_type['logicalType'] == 'decimal':
            import_types.add('decimal.Decimal')
            return 'decimal.Decimal'
        elif avro_type['logicalType'] == 'date':
            import_types.add('datetime.date')
            return 'datetime.date'
        elif avro_type['logicalType'] == 'time-millis':
            import_types.add('datetime.time')
            return 'datetime.time'
        elif avro_type['logicalType'] == 'time-micros':
            import_types.add('datetime.time')
            return 'datetime.time'
        elif avro_type['logicalType'] == 'timestamp-millis':
            import_types.add('datetime.datetime')
            return 'datetime.datetime'
        elif avro_type['logicalType'] == 'timestamp-micros':
            import_types.add('datetime.datetime')
            return 'datetime.datetime'
        elif avro_type['logicalType'] == 'duration':
            import_types.add('datetime.timedelta')
            return 'datetime.timedelta'
        return 'typing.Any'

    def convert_avro_type_to_python(self, avro_type: Union[str, Dict, List], parent_package: str, import_types: set) -> str:
        """Converts Avro type to Python type"""
        if isinstance(avro_type, str):
            is_primitive, mapped_type = self.map_plain_type_reference_to_python(parent_package, avro_type)
            if not is_primitive:
                import_types.add(mapped_type)
                return self.pascal_type_name(mapped_type)
            return mapped_type
        elif isinstance(avro_type, list):
            if is_generic_avro_type(avro_type):
                return 'typing.Any'
            if is_type_with_alternate(avro_type):
                return self.convert_avro_type_to_python(strip_alternate_type(avro_type), parent_package, import_types)
            non_null_types = [t for t in avro_type if t != 'null']
            if len(non_null_types) == 1:
                t = self.convert_avro_type_to_python(non_null_types[0], parent_package, import_types)
                if 'null' in avro_type:
                    return f'typing.Optional[{t}]'
                else:
                    return t
            else:
                return f"typing.Union[{', '.join(self.convert_avro_type_to_python(t, parent_package, import_types) for t in non_null_types)}]"
        elif isinstance(avro_type, dict):
            if avro_type['type'] == 'record':
                class_ref = self.generate_class(avro_type, parent_package, write_file=True)
                import_types.add(class_ref)
                return self.strip_package_from_fully_qualified_name(class_ref)
            elif avro_type['type'] == 'enum':
                enum_ref = self.generate_enum(avro_type, parent_package, write_file=True)
                import_types.add(enum_ref)
                return self.strip_package_from_fully_qualified_name(enum_ref)
            elif avro_type['type'] == 'array':
                return f"typing.List[{self.convert_avro_type_to_python(avro_type['items'], parent_package, import_types)}]"
            elif avro_type['type'] == 'map':
                return f"typing.Dict[str,{self.convert_avro_type_to_python(avro_type['values'], parent_package, import_types)}]"
            elif 'logicalType' in avro_type:
                return self.convert_logical_type_to_python(avro_type, import_types)
            return self.convert_avro_type_to_python(avro_type['type'], parent_package, import_types)
        return 'typing.Any'

    # pylint: disable=eval-used
    def init_field_value(self, field_type: str, field_name: str, field_is_enum: bool, field_ref: str, enum_types: List[str]):
        """ Initialize the field value based on its type. """
        if field_type == "typing.Any":
            return field_ref
        elif field_type in ['datetime.datetime', 'datetime.date', 'datetime.time', 'datetime.timedelta']:
            return f"{field_ref}"
        elif field_type in ['int', 'str', 'float', 'bool', 'bytes', 'Decimal']:
            return f"{field_type}({field_ref})"
        elif field_type.startswith("typing.List["):
            inner_type = get_typing_args_from_string(field_type)[0]
            return f"{field_ref} if isinstance({field_ref}, list) else [{self.init_field_value(inner_type, field_name, field_is_enum, 'v', enum_types)} for v in {field_ref}] if {field_ref} else None"
        elif field_type.startswith("typing.Dict["):
            inner_type = get_typing_args_from_string(field_type)[1]
            return f"{field_ref} if isinstance({field_ref}, dict) else {{k: {self.init_field_value(inner_type, field_name, field_is_enum, 'v', enum_types)} for k, v in {field_ref}.items()}} if {field_ref} else None"
        elif field_type.startswith("typing.Optional["):
            inner_type = get_typing_args_from_string(field_type)[0]
            return self.init_field_value(inner_type, field_name, field_is_enum, field_ref, enum_types) + ' if ' + field_ref + ' else None'
        elif field_type.startswith("typing.Union["):
            return self.init_field_value_from_union(get_typing_args_from_string(field_type), field_name, field_ref, enum_types)
        elif field_is_enum or field_type in enum_types:
            return f"{field_type}({field_ref})"
        else:
            return f"{field_ref} if isinstance({field_ref}, {field_type}) else {field_type}.from_serializer_dict({field_ref}) if {field_ref} else None"

    def init_field_value_from_union(self, union_args: List[str], field_name, field_ref, enum_types):
        """Initialize the field value based on the Union type."""
        init_statements = []
        for field_union_type in union_args:
            init_statements.append(
                f"{self.init_field_value(field_union_type, field_name, field_union_type in enum_types, field_ref, enum_types)} if isinstance({field_ref}, {field_union_type}) else")
        return ' '.join(init_statements) + ' None'

    def init_fields(self, fields: List[Dict[str, Any]], enum_types: List[str]) -> str:
        """Initialize the fields of a class."""
        init_statements = []
        for field in fields:
            if field['is_enum'] or field['type'] in enum_types or field['is_primitive']:
                init_statements.append(
                    f"self.{field['name']}={self.init_field_value(field['type'], field['name'], field['is_enum'], 'self.'+field['name'], enum_types)}")
            else:
                init_statements.append(f"value_{field['name']} = self.{field['name']}")
                init_statements.append(
                    f"self.{field['name']} = {self.init_field_value(field['type'], field['name'], field['is_enum'], 'value_'+field['name'], enum_types)}")
        return '\n'.join(init_statements)

    def generate_class(self, avro_schema: Dict, parent_package: str, write_file: bool) -> str:
        """
        Generates a Python data class from an Avro record schema

        Args:
            avro_schema (Dict): Avro record schema
            parent_package (str): Parent package
            write_file (bool): Write the class to a file

        Returns:
            str: Python fully qualified class name
        """

        import_types: Set[str] = set()
        class_name = self.python_type_from_avro_type(avro_schema['name'])
        package_name = self.python_package_from_avro_type(avro_schema.get('namespace', parent_package), avro_schema['name'])
        python_qualified_name = self.python_fully_qualified_name_from_avro_type(avro_schema.get('namespace', parent_package), avro_schema['name'])
        if python_qualified_name in self.generated_types:
            return python_qualified_name

        fields = [{
            'definition': self.generate_field(field, avro_schema.get('namespace', parent_package), import_types),
            'docstring': self.generate_field_docstring(field, avro_schema.get('namespace', parent_package))
        } for field in avro_schema.get('fields', [])]
        fields = [{
            'name': self.safe_name(field['definition']['name']),
            'original_name': field['definition']['name'],
            'type': field['definition']['type'],
            'is_primitive': field['definition']['is_primitive'],
            'is_enum': field['definition']['is_enum'],
            'docstring': field['docstring'],
            'test_value': self.generate_test_value(field),
        } for field in fields]

        # we are including a copy of the avro schema of this type. Since that may
        # depend on other types, we need to inline all references to other types
        # into this schema. We use deepcopy to avoid mutating the original schema
        # which may be shared with type_dict entries
        local_avro_schema = inline_avro_references(copy.deepcopy(avro_schema), self.type_dict, '')
        avro_schema_json = json.dumps(local_avro_schema).replace('\\"', '\'').replace('"', '\\"')
        enum_types = []
        for import_type in import_types:
            if import_type in self.generated_types and self.generated_types[import_type] == "enum":
                enum_types.append(self.strip_package_from_fully_qualified_name(import_type))

        class_definition = process_template(
            "avrotopython/dataclass_core.jinja",
            class_name=class_name,
            docstring=avro_schema.get('doc', '').strip() if 'doc' in avro_schema else f'A {class_name} record.',
            fields=fields,
            import_types=import_types,
            base_package=self.base_package,
            avro_annotation=self.avro_annotation,
            dataclasses_json_annotation=self.dataclasses_json_annotation,
            avro_schema_json=avro_schema_json,
            init_fields=self.init_fields(fields, enum_types),
        )

        if write_file:
            self.write_to_file(package_name, class_name, class_definition)
            self.generate_test_class(package_name, class_name, fields, import_types)
        self.generated_types[python_qualified_name] = 'class'
        return python_qualified_name

    def generate_enum(self, avro_schema: Dict, parent_package: str, write_file: bool) -> str:
        """
        Generates a Python enum from an Avro enum schema

        Args:
            avro_schema (Dict): Avro enum schema
            parent_package (str): Parent package
            write_file (bool): Write the enum to a file

        Returns:
            str: Python fully qualified enum name
        """

        class_name = self.python_type_from_avro_type(avro_schema['name'])
        package_name = self.python_package_from_avro_type(avro_schema.get('namespace', parent_package), avro_schema['name'])
        python_qualified_name = self.python_fully_qualified_name_from_avro_type(avro_schema.get('namespace', parent_package), avro_schema['name'])
        if python_qualified_name in self.generated_types:
            return python_qualified_name

        symbols = [symbol if not is_python_reserved_word(
            symbol) else symbol + "_" for symbol in avro_schema.get('symbols', [])]
        ordinals =  avro_schema.get('ordinals', {})

        enum_definition = process_template(
            "avrotopython/enum_core.jinja",
            class_name=class_name,
            docstring=avro_schema.get('doc', '').strip(
            ) if 'doc' in avro_schema else f'A {class_name} enum.',
            symbols=symbols,
            ordinals=ordinals
        )

        if write_file:
            self.write_to_file(package_name, class_name, enum_definition)
            self.generate_test_enum(package_name, class_name, symbols)
        self.generated_types[python_qualified_name] = 'enum'
        return python_qualified_name

    def generate_test_class(self, package_name: str, class_name: str, fields: List[Dict[str, str]], import_types: Set[str]) -> None:
        """Generates a unit test class for a Python data class"""
        test_class_name = f"Test_{class_name}"
        tests_package_name = "test_"+package_name.replace('.', '_').lower()
        test_class_definition = process_template(
            "avrotopython/test_class.jinja",
            package_name=package_name,
            class_name=class_name,
            test_class_name=test_class_name,
            fields=fields,
            avro_annotation=self.avro_annotation,
            import_types=import_types
        )

        base_dir = os.path.join(self.output_dir, "tests")
        test_file_path = os.path.join(base_dir, f"{tests_package_name.replace('.', '_').lower()}.py")
        if not os.path.exists(os.path.dirname(test_file_path)):
            os.makedirs(os.path.dirname(test_file_path), exist_ok=True)
        with open(test_file_path, 'w', encoding='utf-8') as file:
            file.write(test_class_definition)

    def generate_test_enum(self, package_name: str, class_name: str, symbols: List[str]) -> None:
        """Generates a unit test class for a Python enum"""
        test_class_name = f"Test_{class_name}"
        tests_package_name = "test_"+package_name.replace('.', '_').lower()
        test_class_definition = process_template(
            "avrotopython/test_enum.jinja",
            package_name=package_name,
            class_name=class_name,
            test_class_name=test_class_name,
            symbols=symbols
        )
        base_dir = os.path.join(self.output_dir, "tests")
        test_file_path = os.path.join(base_dir, f"{tests_package_name.replace('.', '_').lower()}.py")
        if not os.path.exists(os.path.dirname(test_file_path)):
            os.makedirs(os.path.dirname(test_file_path), exist_ok=True)
        with open(test_file_path, 'w', encoding='utf-8') as file:
            file.write(test_class_definition)

    def generate_test_value(self, field: Dict) -> Any:
        """Generates a test value for a given field"""
        field_type = field['definition']['type']

        def generate_value(field_type: str):
            test_values = {
                'str': chr(39)+''.join([chr(random.randint(97, 122)) for _ in range(0, 20)])+chr(39),
                'bool': str(random.choice([True, False])),
                'int': f'int({random.randint(0, 100)})',
                'float': f'float({random.uniform(0, 100)})',
                'bytes': 'b"test_bytes"',
                'None': 'None',
                'datetime.date': random.choice(['datetime.date.today()', 'datetime.date(2021, 1, 1)']),
                'datetime.datetime': 'datetime.datetime.now(datetime.timezone.utc)',
                'datetime.time': 'datetime.datetime.now(datetime.timezone.utc).time()',
                'decimal.Decimal': f'decimal.Decimal("{random.randint(0, 100)}.{random.randint(0, 100)}")',
                'datetime.timedelta': 'datetime.timedelta(days=1)',
                'typing.Any': '{"test": "test"}'
            }

            def resolve(field_type: str) -> str:
                # Regex pattern to find the inner type
                pattern = re.compile(r'^(?:typing\.)*(Optional|List|Dict|Union)\[(.+)\]$')

                match = pattern.match(field_type)
                if not match:
                    return field_type

                outer_type, inner_type = match.groups()

                if outer_type == 'Optional':
                    return inner_type
                elif outer_type == 'List':
                    return resolve(inner_type)
                elif outer_type == 'Dict':
                    # For Dict, only return the value type
                    _, value_type = inner_type.split(',', 1)
                    return resolve(value_type.strip())
                elif outer_type == 'Union':
                    first_type = inner_type.split(',', 1)[0]
                    return resolve(first_type.strip())

                return field_type

            if field_type.startswith('typing.Optional['):
                field_type = resolve(field_type)

            if field_type.startswith('typing.List['):
                field_type = resolve(field_type)
                array_range = random.randint(1, 5)
                return f"[{', '.join([generate_value(field_type) for _ in range(array_range)])}]"
            elif field_type.startswith('typing.Dict['):
                field_type = resolve(field_type)
                dict_range = random.randint(1, 5)
                dict_data = {}
                for _ in range(dict_range):
                    dict_data[''.join([chr(random.randint(97, 122)) for _ in range(
                        0, 20)])] = generate_value(field_type)
                return f"{{{', '.join([chr(39)+key+chr(39)+f': {value}' for key, value in dict_data.items()])}}}"
            elif field_type.startswith('typing.Union['):
                field_type = resolve(field_type)
                return generate_value(field_type)
            return test_values.get(field_type, 'Test_'+field_type + '.create_instance()')

        return generate_value(field_type)

    def generate_field(self, field: Dict, parent_package: str, import_types: set) -> Any:
        """Generates a field for a Python data class"""
        field_type = self.convert_avro_type_to_python(field['type'], parent_package, import_types)
        field_name = field['name']
        return {
            'name': field_name,
            'type': field_type,
            'is_primitive': self.is_python_primitive(field_type) or self.is_python_typing_struct(field_type),
            'is_enum': field_type in self.generated_types and self.generated_types[field_type] == 'enum'
        }

    def generate_field_docstring(self, field: Dict, parent_package: str) -> str:
        """Generates a field docstring for a Python data class"""
        field_type = self.convert_avro_type_to_python(field['type'], parent_package, set())
        field_name = self.safe_name(field['name'])
        field_doc = field.get('doc', '').strip()
        if is_python_reserved_word(field_name):
            field_name += "_"
        field_docstring = f"{field_name} ({field_type}): {field_doc}"
        return field_docstring

    def write_to_file(self, package: str, class_name: str, python_code: str):
        """
        Writes a Python class to a file

        Args:
            package (str): Python package
            class_name (str): Python class name
            python_code (str): Python class definition
        """

        # the containing directory is the parent package
        parent_package_name = '.'.join(package.split('.')[:-1])
        parent_package_path = os.sep.join(parent_package_name.split('.')).lower()
        directory_path = os.path.join(self.output_dir, "src", parent_package_path)
        if not os.path.exists(directory_path):
            os.makedirs(directory_path, exist_ok=True)
        file_path = os.path.join(directory_path, f"{class_name.lower()}.py")

        with open(file_path, 'w', encoding='utf-8') as file:
            file.write(python_code)

    def write_init_files(self):
        """Writes __init__.py files to the output directories"""

        def organize_generated_types():
            """
            Organizes the generated_types into a tree structure.
            
            For a fully qualified name like 'address.example.com.record.Record':
            - The package.module path is 'address.example.com.record'
            - The class name is 'Record'
            - The module file is 'record.py' in directory 'address/example/com/'
            
            The tree structure should navigate through package parts up to the parent directory,
            then store the module name (last package part) as the key with class name as value:
            {
              'address': {
                'example': {
                  'com': {
                    'record': 'Record'  # module 'record.py' -> class 'Record'
                  }
                }
              }
            }
            """
            generated_types_tree = {}
            for generated_type, _ in self.generated_types.items():
                parts = generated_type.split('.')
                if len(parts) < 2:
                    continue  # Need at least module.Class
                    
                class_name = parts[-1]  # Last part is the class name
                module_name = parts[-2]  # Second-to-last is the module name
                package_parts = parts[:-2]  # Everything before module and class is the package path
                
                # Navigate through the package hierarchy
                current_node = generated_types_tree
                for part in package_parts:
                    if part not in current_node:
                        current_node[part] = {}
                    current_node = current_node[part]
                
                # Store module -> class mapping at this level
                current_node[module_name] = class_name
                
            return generated_types_tree

        def collect_class_names(node):
            """
            Recursively collect all class names from a tree node.
            Returns a list of class names (PascalCase) available in this package level.
            """
            class_names = []
            for key, value in node.items():
                if isinstance(value, dict):
                    # Recursively collect from subpackages
                    class_names.extend(collect_class_names(value))
                else:
                    # This is a leaf node with a class name
                    class_names.append(value)
            return class_names

        def write_init_files_recursive(generated_types_tree, current_package: str):
            """
            Writes __init__.py files recursively.
            
            For each package level:
            - Import classes from module files (from .modulename import ClassName)
            - Import classes from subpackages (from .subpackage import ClassName)
            - Re-export all class names in __all__
            """
            import_statements = []
            all_statement = []
            
            for package_or_module_name, content in generated_types_tree.items():
                if isinstance(content, dict):
                    # This is a subpackage - collect all class names from it
                    class_names = collect_class_names(content)
                    if class_names:
                        import_statements.append(f"from .{package_or_module_name} import {', '.join(class_names)}")
                        all_statement.extend([f'"{name}"' for name in class_names])
                    # Recursively write __init__.py for the subpackage
                    write_init_files_recursive(content, current_package + ('.' if current_package else '') + package_or_module_name)
                else:
                    # This is a module file - import the class from it
                    class_name = content
                    import_statements.append(f"from .{package_or_module_name} import {class_name}")
                    all_statement.append(f'"{class_name}"')
            
            if current_package and (import_statements or all_statement):
                package_path = os.path.join(self.output_dir, 'src', current_package.replace('.', os.sep).lower())
                init_file_path = os.path.join(package_path, '__init__.py')
                if not os.path.exists(package_path):
                    os.makedirs(package_path, exist_ok=True)
                with open(init_file_path, 'w', encoding='utf-8') as file:
                    file.write('\n'.join(import_statements) + '\n\n__all__ = [' + ', '.join(all_statement) + ']\n')

        # main function
        write_init_files_recursive(organize_generated_types(), '')

    def write_pyproject_toml(self):
        """Writes pyproject.toml file to the output directory"""
        pyproject_content = process_template(
            "avrotopython/pyproject_toml.jinja",
            package_name=self.base_package.replace('_', '-')
        )
        with open(os.path.join(self.output_dir, 'pyproject.toml'), 'w', encoding='utf-8') as file:
            file.write(pyproject_content)

    def convert_schemas(self, avro_schemas: List, output_dir: str):
        """ Converts Avro schema to Python data classes"""
        self.main_schema = avro_schemas
        self.type_dict = build_flat_type_dict(avro_schemas)
        self.output_dir = output_dir
        if not os.path.exists(self.output_dir):
            os.makedirs(self.output_dir, exist_ok=True)
        for avro_schema in avro_schemas:
            if avro_schema['type'] == 'enum':
                self.generate_enum(
                    avro_schema, self.base_package, write_file=True)
            elif avro_schema['type'] == 'record':
                self.generate_class(avro_schema, self.base_package, write_file=True)
        self.write_init_files()
        self.write_pyproject_toml()

    def convert(self, avro_schema_path: str, output_dir: str):
        """Converts Avro schema to Python data classes"""
        with open(avro_schema_path, 'r', encoding='utf-8') as file:
            schema = json.load(file)
        if isinstance(schema, dict):
            schema = [schema]
        return self.convert_schemas(schema, output_dir)


def convert_avro_to_python(avro_schema_path, py_file_path, package_name='', dataclasses_json_annotation=False, avro_annotation=False):
    """Converts Avro schema to Python data classes"""
    if not package_name:
        package_name = os.path.splitext(os.path.basename(avro_schema_path))[
            0].lower().replace('-', '_')

    avro_to_python = AvroToPython(
        package_name, dataclasses_json_annotation=dataclasses_json_annotation, avro_annotation=avro_annotation)
    avro_to_python.convert(avro_schema_path, py_file_path)


def convert_avro_schema_to_python(avro_schema, py_file_path, package_name='', dataclasses_json_annotation=False, avro_annotation=False):
    """Converts Avro schema to Python data classes"""
    avro_to_python = AvroToPython(
        package_name, dataclasses_json_annotation=dataclasses_json_annotation, avro_annotation=avro_annotation)
    if isinstance(avro_schema, dict):
        avro_schema = [avro_schema]
    avro_to_python.convert_schemas(avro_schema, py_file_path)



# coding: utf-8
"""
Module to convert Avro schema to CSV schema.
"""

import json
import os

class AvroToCSVSchemaConverter:
    """
    Class to convert Avro schema to CSV schema.
    """

    def __init__(self, avro_schema_path, csv_schema_path):
        """
        Initialize the converter with file paths.

        :param avro_schema_path: Path to the Avro schema file.
        :param csv_schema_path: Path to save the CSV schema file.
        """
        self.avro_schema_path = avro_schema_path
        self.csv_schema_path = csv_schema_path

    def convert(self):
        """
        Convert Avro schema to CSV schema and save to file.
        """
        with open(self.avro_schema_path, 'r', encoding='utf-8') as file:
            avro_schema = json.load(file)

        csv_schema = self.convert_avro_to_csv_schema(avro_schema)

        with open(self.csv_schema_path, 'w', encoding='utf-8') as file:
            json.dump(csv_schema, file, indent=2)

    def convert_avro_to_csv_schema(self, avro_schema):
        """
        Convert an Avro schema to a CSV schema.

        :param avro_schema: Avro schema as a dictionary.
        :return: CSV schema as a dictionary.
        """
        csv_schema = {
            "fields": []
        }

        for field in avro_schema['fields']:
            csv_field = self.convert_avro_field_to_csv_field(field)
            csv_schema['fields'].append(csv_field)

        return csv_schema

    def convert_avro_field_to_csv_field(self, field):
        """
        Convert an Avro field to a CSV field.

        :param field: Avro field as a dictionary.
        :return: CSV field as a dictionary.
        """
        csv_field = {
            "name": field['name'],
            "type": self.convert_avro_type_to_csv_type(field['type'])
        }

        if isinstance(field['type'], dict) and field['type'].get('logicalType') == 'timestamp-millis':
            csv_field['type'] = 'string'
            csv_field['format'] = 'datetime'

        if field.get('default') is not None:
            csv_field['default'] = field['default']

        if isinstance(field['type'], list) and "null" in field['type']:
            csv_field['nullable'] = True

        if 'enum' in field['type']:
            csv_field['enum'] = field['type']['symbols']

        return csv_field

    def convert_avro_type_to_csv_type(self, avro_type):
        """
        Convert an Avro type to a CSV type.

        :param avro_type: Avro type as a string or dictionary.
        :return: CSV type as a string.
        """
        if isinstance(avro_type, list):
            avro_type = [t for t in avro_type if t != "null"][0]

        type_mapping = {
            "string": "string",
            "int": "integer",
            "long": "integer",
            "float": "number",
            "double": "number",
            "boolean": "boolean",
            "bytes": "string",
            "record": "object",
            "array": "array",
            "map": "object",
            "enum": "string"
        }

        if isinstance(avro_type, dict):
            return type_mapping.get(avro_type['type'], "string")

        return type_mapping.get(avro_type, "string")

def convert_avro_to_csv_schema(avro_schema_path, csv_schema_path):
    """
    Convert an Avro schema file to a CSV schema file.

    :param avro_schema_path: Path to the Avro schema file.
    :param csv_schema_path: Path to save the CSV schema file.
    """
    
    if not os.path.exists(avro_schema_path):
        raise FileNotFoundError(f"Avro schema file not found: {avro_schema_path}")
    
    converter = AvroToCSVSchemaConverter(avro_schema_path, csv_schema_path)
    converter.convert()



# pylint: disable=line-too-long

""" StructureToGo class for converting JSON Structure schema to Go structs """

import json
import os
from typing import Any, Dict, List, Set, Union, Optional, cast

from avrotize.common import pascal, render_template

JsonNode = Dict[str, 'JsonNode'] | List['JsonNode'] | str | None

INDENT = '    '


class StructureToGo:
    """ Converts JSON Structure schema to Go structs """

    def __init__(self, base_package: str = '') -> None:
        self.base_package = base_package
        self.output_dir = os.getcwd()
        self.json_annotation = False
        self.avro_annotation = False
        self.package_site = 'github.com'
        self.package_username = 'username'
        self.schema_doc: JsonNode = None
        self.generated_types: Dict[str, str] = {}
        self.generated_structure_types: Dict[str, Dict[str, Union[str, Dict, List]]] = {}
        self.definitions: Dict[str, Any] = {}
        self.schema_registry: Dict[str, Dict] = {}
        self.structs: List[Dict] = []
        self.enums: List[Dict] = []

    def safe_identifier(self, name: str) -> str:
        """Converts a name to a safe Go identifier"""
        reserved_words = [
            'break', 'default', 'func', 'interface', 'select', 'case', 'defer', 'go', 'map', 'struct', 'chan',
            'else', 'goto', 'package', 'switch', 'const', 'fallthrough', 'if', 'range', 'type', 'continue', 'for',
            'import', 'return', 'var',
        ]
        if name in reserved_words:
            return f"{name}_"
        return name

    def go_type_name(self, name: str, namespace: str = '') -> str:
        """Returns a qualified name for a Go struct or enum"""
        if namespace:
            namespace = ''.join([pascal(part) for part in namespace.split('.')])
            return f"{namespace}{pascal(name)}"
        return pascal(name)

    def map_primitive_to_go(self, structure_type: str, is_optional: bool) -> str:
        """Maps JSON Structure primitive types to Go types"""
        optional_mapping = {
            'null': 'interface{}',
            'boolean': '*bool',
            'string': '*string',
            'integer': '*int',
            'number': '*float64',
            'int8': '*int8',
            'uint8': '*uint8',
            'int16': '*int16',
            'uint16': '*uint16',
            'int32': '*int32',
            'uint32': '*uint32',
            'int64': '*int64',
            'uint64': '*uint64',
            'int128': '*big.Int',
            'uint128': '*big.Int',
            'float8': '*float32',
            'float': '*float32',
            'double': '*float64',
            'binary32': '*float32',
            'binary64': '*float64',
            'decimal': '*float64',
            'binary': '[]byte',
            'date': '*time.Time',
            'time': '*time.Time',
            'datetime': '*time.Time',
            'timestamp': '*time.Time',
            'duration': '*time.Duration',
            'uuid': '*string',
            'uri': '*string',
            'jsonpointer': '*string',
            'any': 'interface{}',
        }
        required_mapping = {
            'null': 'interface{}',
            'boolean': 'bool',
            'string': 'string',
            'integer': 'int',
            'number': 'float64',
            'int8': 'int8',
            'uint8': 'uint8',
            'int16': 'int16',
            'uint16': 'uint16',
            'int32': 'int32',
            'uint32': 'uint32',
            'int64': 'int64',
            'uint64': 'uint64',
            'int128': 'big.Int',
            'uint128': 'big.Int',
            'float8': 'float32',
            'float': 'float32',
            'double': 'float64',
            'binary32': 'float32',
            'binary64': 'float64',
            'decimal': 'float64',
            'binary': '[]byte',
            'date': 'time.Time',
            'time': 'time.Time',
            'datetime': 'time.Time',
            'timestamp': 'time.Time',
            'duration': 'time.Duration',
            'uuid': 'string',
            'uri': 'string',
            'jsonpointer': 'string',
            'any': 'interface{}',
        }
        if structure_type in self.generated_types:
            return structure_type
        else:
            return required_mapping.get(structure_type, 'interface{}') if not is_optional else optional_mapping.get(structure_type, 'interface{}')

    def resolve_ref(self, ref: str, context_schema: Optional[Dict] = None) -> Optional[Dict]:
        """ Resolves a $ref to the actual schema definition """
        # Check if it's an absolute URI reference
        if not ref.startswith('#/'):
            if ref in self.schema_registry:
                return self.schema_registry[ref]
            return None

        # Handle fragment-only references
        path = ref[2:].split('/')
        schema = context_schema if context_schema else self.schema_doc

        for part in path:
            if not isinstance(schema, dict) or part not in schema:
                return None
            schema = schema[part]

        return schema

    def register_schema_ids(self, schema: Dict, base_uri: str = '') -> None:
        """ Recursively registers schemas with $id keywords """
        if not isinstance(schema, dict):
            return

        if '$id' in schema:
            schema_id = schema['$id']
            if base_uri and not schema_id.startswith(('http://', 'https://', 'urn:')):
                from urllib.parse import urljoin
                schema_id = urljoin(base_uri, schema_id)
            self.schema_registry[schema_id] = schema
            base_uri = schema_id

        if 'definitions' in schema:
            for def_name, def_schema in schema['definitions'].items():
                if isinstance(def_schema, dict):
                    self.register_schema_ids(def_schema, base_uri)

        if 'properties' in schema:
            for prop_name, prop_schema in schema['properties'].items():
                if isinstance(prop_schema, dict):
                    self.register_schema_ids(prop_schema, base_uri)

        for key in ['items', 'values', 'additionalProperties']:
            if key in schema and isinstance(schema[key], dict):
                self.register_schema_ids(schema[key], base_uri)

    def convert_structure_type_to_go(self, class_name: str, field_name: str, 
                                     structure_type: JsonNode, parent_namespace: str, 
                                     nullable: bool = False) -> str:
        """ Converts JSON Structure type to Go type """
        if isinstance(structure_type, str):
            return self.map_primitive_to_go(structure_type, nullable)
        elif isinstance(structure_type, list):
            # Handle type unions
            non_null_types = [t for t in structure_type if t != 'null']
            has_null = 'null' in structure_type
            if len(non_null_types) == 1:
                inner_type = self.convert_structure_type_to_go(
                    class_name, field_name, non_null_types[0], parent_namespace, has_null)
                return inner_type
            else:
                # Generate union type
                return self.generate_union_class(field_name, structure_type, parent_namespace)
        elif isinstance(structure_type, dict):
            # Handle $ref
            if '$ref' in structure_type:
                ref_schema = self.resolve_ref(structure_type['$ref'], self.schema_doc)
                if ref_schema:
                    ref_path = structure_type['$ref'].split('/')
                    type_name = ref_path[-1]
                    ref_namespace = '.'.join(ref_path[2:-1]) if len(ref_path) > 3 else parent_namespace
                    return self.generate_class_or_choice(ref_schema, ref_namespace, write_file=True, explicit_name=type_name)
                return 'interface{}'

            # Handle enum keyword
            if 'enum' in structure_type:
                return self.generate_enum(structure_type, field_name, parent_namespace, write_file=True)

            # Handle type keyword
            if 'type' not in structure_type:
                return 'interface{}'

            struct_type = structure_type['type']

            # Handle complex types
            if struct_type == 'object':
                return self.generate_class(structure_type, parent_namespace, write_file=True)
            elif struct_type == 'array':
                items_type = self.convert_structure_type_to_go(
                    class_name, field_name+'List', structure_type.get('items', {'type': 'any'}), 
                    parent_namespace, nullable=True)
                if items_type.startswith('*'):
                    return f"[]{items_type[1:]}"
                return f"[]{items_type}"
            elif struct_type == 'set':
                # Go doesn't have a built-in set, use map[T]bool
                items_type = self.convert_structure_type_to_go(
                    class_name, field_name+'Set', structure_type.get('items', {'type': 'any'}), 
                    parent_namespace, nullable=True)
                if items_type.startswith('*'):
                    items_type = items_type[1:]
                return f"map[{items_type}]bool"
            elif struct_type == 'map':
                values_type = self.convert_structure_type_to_go(
                    class_name, field_name+'Map', structure_type.get('values', {'type': 'any'}), 
                    parent_namespace, nullable=True)
                return f"map[string]{values_type}"
            elif struct_type == 'choice':
                return self.generate_choice(structure_type, parent_namespace, write_file=True)
            elif struct_type == 'tuple':
                return self.generate_tuple(structure_type, parent_namespace, write_file=True)
            else:
                return self.convert_structure_type_to_go(class_name, field_name, struct_type, parent_namespace, nullable)
        return 'interface{}'

    def generate_class_or_choice(self, structure_schema: Dict, parent_namespace: str, 
                                 write_file: bool = True, explicit_name: str = '') -> str:
        """ Generates a Class or Choice """
        struct_type = structure_schema.get('type', 'object')
        if struct_type == 'object':
            return self.generate_class(structure_schema, parent_namespace, write_file, explicit_name=explicit_name)
        elif struct_type == 'choice':
            return self.generate_choice(structure_schema, parent_namespace, write_file, explicit_name=explicit_name)
        elif struct_type == 'tuple':
            return self.generate_tuple(structure_schema, parent_namespace, write_file, explicit_name=explicit_name)
        return 'interface{}'

    def generate_class(self, structure_schema: Dict, parent_namespace: str, 
                      write_file: bool, explicit_name: str = '') -> str:
        """ Generates a Go struct from JSON Structure object type """
        class_name = pascal(explicit_name if explicit_name else structure_schema.get('name', 'UnnamedClass'))
        schema_namespace = structure_schema.get('namespace', parent_namespace)
        go_struct_name = self.go_type_name(class_name, schema_namespace)

        if go_struct_name in self.generated_types:
            return go_struct_name

        # Check if this is an abstract type
        is_abstract = structure_schema.get('abstract', False)
        
        # If abstract, generate interface instead
        if is_abstract:
            return self.generate_interface(structure_schema, parent_namespace, write_file, explicit_name)

        self.generated_types[go_struct_name] = "struct"
        self.generated_structure_types[go_struct_name] = structure_schema

        # Handle inheritance ($extends)
        base_interface = None
        base_properties = {}
        base_required = []
        if '$extends' in structure_schema:
            base_ref = structure_schema['$extends']
            base_schema = self.resolve_ref(base_ref, self.schema_doc)
            if base_schema:
                ref_path = base_ref.split('/')
                base_name = ref_path[-1]
                ref_namespace = '.'.join(ref_path[2:-1]) if len(ref_path) > 3 else parent_namespace
                base_interface = self.generate_class(base_schema, ref_namespace, write_file=True, explicit_name=base_name)
                # Collect base properties to include in the concrete type
                base_properties = base_schema.get('properties', {})
                base_required = base_schema.get('required', [])

        # Generate properties - merge base properties with current properties
        properties = {**base_properties, **structure_schema.get('properties', {})}
        required_props = base_required + structure_schema.get('required', [])

        fields = []
        for prop_name, prop_schema in properties.items():
            is_required = prop_name in required_props if not isinstance(required_props, list) or \
                         len(required_props) == 0 or not isinstance(required_props[0], list) else \
                         any(prop_name in req_set for req_set in required_props)
            
            field_type = self.convert_structure_type_to_go(
                class_name, prop_name, prop_schema, schema_namespace, nullable=not is_required)
            
            # Add nullable marker if not required and not already nullable
            if not is_required and not field_type.startswith('*') and not field_type.startswith('[') and not field_type.startswith('map[') and field_type != 'interface{}':
                field_type = f'*{field_type}'
            
            fields.append({
                'name': pascal(prop_name),
                'type': field_type,
                'original_name': prop_name
            })

        # Get imports needed
        imports = self.get_imports_for_fields([f['type'] for f in fields])
        
        # Generate Avro schema if avro_annotation is enabled
        avro_schema_str = None
        if self.avro_annotation:
            try:
                from avrotize.jstructtoavro import JsonStructureToAvro
                converter = JsonStructureToAvro()
                avro_schema = converter.convert(structure_schema)
                avro_schema_str = json.dumps(avro_schema)
            except Exception as e:
                # If conversion fails, log but continue without Avro schema
                print(f"Warning: Failed to generate Avro schema for {go_struct_name}: {e}")

        context = {
            'doc': structure_schema.get('description', structure_schema.get('doc', class_name)),
            'struct_name': go_struct_name,
            'fields': fields,
            'imports': imports,
            'json_annotation': self.json_annotation,
            'avro_annotation': self.avro_annotation,
            'avro_schema': avro_schema_str,
            'base_package': self.base_package,
            'base_interface': base_interface,
            'referenced_packages': set(),
        }

        pkg_dir = os.path.join(self.output_dir, 'pkg', self.base_package)
        if not os.path.exists(pkg_dir):
            os.makedirs(pkg_dir, exist_ok=True)
        file_name = os.path.join(pkg_dir, f"{go_struct_name}.go")
        render_template('structuretogo/go_struct.jinja', file_name, **context)

        self.structs.append({
            'name': go_struct_name,
            'fields': fields,
        })

        self.generate_unit_test('struct', go_struct_name, fields)

        return go_struct_name

    def generate_enum(self, structure_schema: Dict, field_name: str, parent_namespace: str, 
                     write_file: bool) -> str:
        """ Generates a Go enum from JSON Structure enum """
        enum_name = pascal(structure_schema.get('name', field_name + 'Enum'))
        schema_namespace = structure_schema.get('namespace', parent_namespace)
        go_enum_name = self.go_type_name(enum_name, schema_namespace)

        if go_enum_name in self.generated_types:
            return go_enum_name

        self.generated_types[go_enum_name] = "enum"
        self.generated_structure_types[go_enum_name] = structure_schema

        symbols = structure_schema.get('enum', [])
        
        # Determine base type
        base_type = structure_schema.get('type', 'string')
        go_base_type = self.map_primitive_to_go(base_type, False)

        context = {
            'doc': structure_schema.get('description', structure_schema.get('doc', enum_name)),
            'enum_name': go_enum_name,
            'symbols': symbols,
            'base_type': go_base_type,
            'base_package': self.base_package,
        }

        pkg_dir = os.path.join(self.output_dir, 'pkg', self.base_package)
        if not os.path.exists(pkg_dir):
            os.makedirs(pkg_dir, exist_ok=True)
        file_name = os.path.join(pkg_dir, f"{go_enum_name}.go")
        render_template('structuretogo/go_enum.jinja', file_name, **context)

        self.enums.append({
            'name': go_enum_name,
            'symbols': symbols,
        })

        self.generate_unit_test('enum', go_enum_name, symbols)

        return go_enum_name

    def generate_interface(self, structure_schema: Dict, parent_namespace: str,
                          write_file: bool, explicit_name: str = '') -> str:
        """ Generates a Go interface from JSON Structure abstract type """
        interface_name = pascal(explicit_name if explicit_name else structure_schema.get('name', 'UnnamedInterface'))
        schema_namespace = structure_schema.get('namespace', parent_namespace)
        go_interface_name = self.go_type_name(interface_name, schema_namespace)

        if go_interface_name in self.generated_types:
            return go_interface_name

        self.generated_types[go_interface_name] = "interface"
        self.generated_structure_types[go_interface_name] = structure_schema

        # Get properties to define getter methods
        properties = structure_schema.get('properties', {})
        required_props = structure_schema.get('required', [])
        
        methods = []
        for prop_name, prop_schema in properties.items():
            go_prop_name = pascal(prop_name)
            is_required = prop_name in required_props if not isinstance(required_props, list) or \
                         len(required_props) == 0 or not isinstance(required_props[0], list) else \
                         any(prop_name in req_set for req_set in required_props)
            
            field_type = self.convert_structure_type_to_go(
                interface_name, prop_name, prop_schema, schema_namespace, nullable=not is_required)
            
            # Add nullable marker if not required and not already nullable
            if not is_required and not field_type.startswith('*') and not field_type.startswith('[') and not field_type.startswith('map[') and field_type != 'interface{}':
                field_type = f'*{field_type}'
            
            # Generate getter method
            methods.append({
                'name': f'Get{go_prop_name}',
                'return_type': field_type,
                'doc': f'Get{go_prop_name} returns the {prop_name} field'
            })
            
            # Generate setter method
            methods.append({
                'name': f'Set{go_prop_name}',
                'return_type': '',
                'param_type': field_type,
                'param_name': prop_name,
                'doc': f'Set{go_prop_name} sets the {prop_name} field'
            })

        context = {
            'doc': structure_schema.get('description', structure_schema.get('doc', interface_name)),
            'interface_name': go_interface_name,
            'methods': methods,
            'base_package': self.base_package,
        }

        pkg_dir = os.path.join(self.output_dir, 'pkg', self.base_package)
        if not os.path.exists(pkg_dir):
            os.makedirs(pkg_dir, exist_ok=True)
        file_name = os.path.join(pkg_dir, f"{go_interface_name}.go")
        render_template('structuretogo/go_interface.jinja', file_name, **context)

        return go_interface_name

    def generate_choice(self, structure_schema: Dict, parent_namespace: str, 
                       write_file: bool, explicit_name: str = '') -> str:
        """ Generates a choice/union type """
        # For simplicity, generate as interface{} for now
        # A complete implementation would generate proper union types
        return 'interface{}'

    def generate_tuple(self, structure_schema: Dict, parent_namespace: str, 
                      write_file: bool, explicit_name: str = '') -> str:
        """ Generates a tuple type """
        # For simplicity, generate as interface{} for now
        return 'interface{}'

    def generate_union_class(self, field_name: str, structure_types: List, parent_namespace: str) -> str:
        """ Generates a union type """
        # For simplicity, generate as interface{} for now
        return 'interface{}'

    def get_imports_for_fields(self, types: List[str]) -> Set[str]:
        """Collects necessary imports based on the Go types"""
        imports = set()
        for field_type in types:
            if "time.Time" in field_type or "time.Duration" in field_type:
                imports.add("time")
            if "big.Int" in field_type:
                imports.add("math/big")
        return imports

    def random_value(self, go_type: str) -> str:
        """Generates a random value for a given Go type"""
        import random
        import string

        is_optional = False
        if go_type.startswith('*'):
            is_optional = True
            go_type = go_type[1:]

        if go_type == 'string':
            v = '"' + ''.join(random.choices(string.ascii_letters + string.digits, k=10)) + '"'
        elif go_type == 'bool':
            v = 'true' if random.choice([True, False]) else 'false'
        elif go_type == 'int':
            v = str(random.randint(-100, 100))
        elif go_type == 'int8':
            v = f'int8({random.randint(-100, 100)})'
        elif go_type == 'int16':
            v = f'int16({random.randint(-100, 100)})'
        elif go_type == 'int32':
            v = f'int32({random.randint(-100, 100)})'
        elif go_type == 'int64':
            v = f'int64({random.randint(-100, 100)})'
        elif go_type == 'uint':
            v = f'uint({random.randint(0, 200)})'
        elif go_type == 'uint8':
            v = f'uint8({random.randint(0, 200)})'
        elif go_type == 'uint16':
            v = f'uint16({random.randint(0, 200)})'
        elif go_type == 'uint32':
            v = f'uint32({random.randint(0, 200)})'
        elif go_type == 'uint64':
            v = f'uint64({random.randint(0, 200)})'
        elif go_type == 'float32':
            v = f'float32({random.uniform(-100, 100)})'
        elif go_type == 'float64':
            v = f'float64({random.uniform(-100, 100)})'
        elif go_type == '[]byte':
            v = '[]byte("' + ''.join(random.choices(string.ascii_letters + string.digits, k=10)) + '")'
        elif go_type.startswith('[]'):
            inner_type = go_type[2:]
            v = f'{go_type}{{{self.random_value(inner_type)}}}'
        elif go_type.startswith('map['):
            # Extract value type from map[KeyType]ValueType
            value_type = go_type.split(']', 1)[1]
            v = f'{go_type}{{"key": {self.random_value(value_type)}}}'
        elif go_type in self.generated_types:
            v = f'CreateInstance{go_type}()'
        elif go_type == 'interface{}':
            v = 'nil'
        else:
            v = 'nil'

        if is_optional and v != 'nil':
            # Create a helper function to get pointer with proper type
            return f'func() *{go_type} {{ v := {v}; return &v }}()'
        return v

    def generate_helpers(self) -> None:
        """Generates helper functions for initializing structs with random values"""
        context = {
            'structs': self.structs,
            'enums': self.enums,
            'base_package': self.base_package,
        }
        for struct in context['structs']:
            for field in struct['fields']:
                if 'value' not in field:
                    field['value'] = self.random_value(field['type'])
        helpers_file_name = os.path.join(self.output_dir, 'pkg', self.base_package, f"{self.base_package}_helpers.go")
        render_template('structuretogo/go_helpers.jinja', helpers_file_name, **context)

    def generate_unit_test(self, kind: str, name: str, fields: Any):
        """Generates unit tests for Go struct or enum"""
        context = {
            'struct_name': name,
            'fields': fields,
            'kind': kind,
            'base_package': self.base_package,
            'package_site': self.package_site,
            'package_username': self.package_username,
            'json_annotation': self.json_annotation,
            'avro_annotation': self.avro_annotation
        }

        pkg_dir = os.path.join(self.output_dir, 'pkg', self.base_package)
        if not os.path.exists(pkg_dir):
            os.makedirs(pkg_dir, exist_ok=True)
        test_file_name = os.path.join(pkg_dir, f"{name}_test.go")
        render_template('structuretogo/go_test.jinja', test_file_name, **context)

    def write_go_mod_file(self):
        """Writes the go.mod file for the Go project"""
        go_mod_content = ""
        go_mod_content += "module " + self.package_site + "/" + self.package_username + "/" + self.base_package + "\n\n"
        go_mod_content += "go 1.21\n\n"
        if self.avro_annotation:
            go_mod_content += "require (\n"
            go_mod_content += "    github.com/hamba/avro/v2 v2.27.0\n"
            go_mod_content += ")\n"

        go_mod_path = os.path.join(self.output_dir, "go.mod")
        with open(go_mod_path, 'w', encoding='utf-8') as file:
            file.write(go_mod_content)

    def write_modname_go_file(self):
        """Writes the modname.go file for the Go project"""
        modname_go_content = ""
        modname_go_content += "package " + self.base_package + "\n\n"
        modname_go_content += "const ModName = \"" + self.base_package + "\"\n"

        modname_go_path = os.path.join(self.output_dir, 'pkg', self.base_package, "module.go")
        with open(modname_go_path, 'w', encoding='utf-8') as file:
            file.write(modname_go_content)

    def process_definitions(self, definitions: Dict, namespace_path: str) -> None:
        """ Processes the definitions section recursively """
        for name, definition in definitions.items():
            if isinstance(definition, dict):
                if 'type' in definition or 'enum' in definition:
                    # This is a type definition
                    current_namespace = namespace_path
                    # Check if this type was already generated
                    check_name = self.go_type_name(name, current_namespace)
                    if check_name not in self.generated_types:
                        if 'enum' in definition:
                            self.generate_enum(definition, name, current_namespace, write_file=True)
                        else:
                            self.generate_class_or_choice(definition, current_namespace, write_file=True, explicit_name=name)
                else:
                    # This is a nested namespace
                    new_namespace = f"{namespace_path}.{name}" if namespace_path else name
                    self.process_definitions(definition, new_namespace)

    def convert_schema(self, schema: JsonNode, output_dir: str):
        """Converts JSON Structure schema to Go"""
        if not isinstance(schema, list):
            schema = [schema]
        if not os.path.exists(output_dir):
            os.makedirs(output_dir, exist_ok=True)
        self.output_dir = output_dir

        self.structs = []
        self.enums = []

        # Register all schemas with $id keywords for cross-references
        for structure_schema in (s for s in schema if isinstance(s, dict)):
            self.register_schema_ids(structure_schema)

        for structure_schema in (s for s in schema if isinstance(s, dict)):
            self.schema_doc = structure_schema
            
            # Store definitions for later use
            if 'definitions' in structure_schema:
                self.definitions = structure_schema['definitions']

            # Process root type
            if 'enum' in structure_schema:
                self.generate_enum(structure_schema, structure_schema.get('name', 'Enum'), 
                                 structure_schema.get('namespace', ''), write_file=True)
            elif 'type' in structure_schema:
                self.generate_class_or_choice(structure_schema, '', write_file=True)
            elif '$root' in structure_schema:
                root_ref = structure_schema['$root']
                root_schema = self.resolve_ref(root_ref, structure_schema)
                if root_schema:
                    ref_path = root_ref.split('/')
                    type_name = ref_path[-1]
                    ref_namespace = '.'.join(ref_path[2:-1]) if len(ref_path) > 3 else ''
                    self.generate_class_or_choice(root_schema, ref_namespace, write_file=True, explicit_name=type_name)

            # Process remaining definitions
            if 'definitions' in structure_schema:
                self.process_definitions(self.definitions, '')

        self.write_go_mod_file()
        self.write_modname_go_file()
        self.generate_helpers()

    def convert(self, structure_schema_path: str, output_dir: str):
        """Converts JSON Structure schema to Go"""
        if not self.base_package:
            self.base_package = os.path.splitext(os.path.basename(structure_schema_path))[0].replace('-', '_').lower()

        with open(structure_schema_path, 'r', encoding='utf-8') as file:
            schema = json.load(file)
        self.convert_schema(schema, output_dir)


def convert_structure_to_go(structure_schema_path: str, go_file_path: str, package_name: str = '',
                            json_annotation: bool = False, avro_annotation: bool = False,
                            package_site: str = 'github.com', package_username: str = 'username'):
    """Converts JSON Structure schema to Go structs

    Args:
        structure_schema_path (str): JSON Structure input schema path
        go_file_path (str): Output Go directory path
        package_name (str): Base package name
        json_annotation (bool): Include JSON annotations
        avro_annotation (bool): Include Avro annotations
        package_site (str): Package site for Go module
        package_username (str): Package username for Go module
    """
    if not package_name:
        package_name = os.path.splitext(os.path.basename(structure_schema_path))[0].replace('-', '_').lower()

    structuretogo = StructureToGo(package_name)
    structuretogo.json_annotation = json_annotation
    structuretogo.avro_annotation = avro_annotation
    structuretogo.package_site = package_site
    structuretogo.package_username = package_username
    structuretogo.convert(structure_schema_path, go_file_path)


def convert_structure_schema_to_go(structure_schema: JsonNode, output_dir: str, package_name: str = '',
                                   json_annotation: bool = False, avro_annotation: bool = False,
                                   package_site: str = 'github.com', package_username: str = 'username'):
    """Converts JSON Structure schema to Go structs

    Args:
        structure_schema (JsonNode): JSON Structure schema as a dictionary or list of dictionaries
        output_dir (str): Output directory path
        package_name (str): Base package name
        json_annotation (bool): Include JSON annotations
        avro_annotation (bool): Include Avro annotations
        package_site (str): Package site for Go module
        package_username (str): Package username for Go module
    """
    structuretogo = StructureToGo(package_name)
    structuretogo.json_annotation = json_annotation
    structuretogo.avro_annotation = avro_annotation
    structuretogo.package_site = package_site
    structuretogo.package_username = package_username
    structuretogo.convert_schema(structure_schema, output_dir)
